unit xtree2_unit;
{
   text element record contains:
 - properties....
   .position (x,y) //bytes
   .type
   ...
 - links/undo info
   .parent         //word
   .child          //word
   .next           //word
   .previous       //word
   .Cstep          //byte
   .Dstep          //byte

 modifications:
 1.
 - undo stack removed
 instead:
 - each element is attached a
    - creation step number (Cstep)
    - deletion step number (Dstep)

 2.
 deleted element is not removed from tree but is skipped.

 Edstep (edit action) counter counts 1,2,3..maxStep,1,2,3... circular
 Edbase is 1st count
 overflow if EdStep = Edbase

 ----------------------------------------------------------------------------
 Edit step counters Edbase, Edstep
 - initially Edbase=0, Edstep=0

 incrementing Edstep counter
 ---------------------------
 - if Edstep=0 then set Edstep=1, Edbase=1
 - if Edstep=maxUndo then Edstep=1 else increment Edstep
 - scan elements for Cstep=Edstep
   if true, set element Cstep=0   (placing element outside undo range)
 - scan elements for Dstep=Edstep
   if true, destroy element

 destroy element
 ----------------
 - set element type to free to enable re-use
 - set Cstep=0, Dstep=0.
 do not destroy children (may be linked to other element)

 decrement Edstep counter
 ------------------------
 - if Edstep=Edbase then Edstep=0, Edbase=0 else decrement Edstep
 -----------------------------------------------------------------------------

 Operations:
 1. Adding an element
    - increment edStep counter
    - link element in tree
    - set element Cstep = Edstep

 2. Deleting an element
    - increment edStep counter
    - set element Dstep = EdStep

 3. Replacing an element
    - increment EdStep counter
    - set element Dstep = EdStep + $80      ($80 is replace flag)
    - link new element as next
    - set new element Cstep = EdStep + $80  (..)
    - link children to new element

if linking, increment of Edstep is blocked.

Undo :
 scan elements

1U  undo added element
    - if element Cstep = Edstep then
      - if (Cstep and $80) = 0 then destroy children
      - destroy element
    - decrement EdStep counter

2U  undo deleted element
    - if element Dstep = EdStep then Dstep = 0
    - if (Dstep and $80 = $80) then link children
    - decrement Edstep counter

3U  see 1U,2U

 ------------------------------------------------------------------------------

   operations      description
   ----------      --------------
   placeA          place after
   placeB          place before
   placeChildA     place as last child
   placeChildB     place as first child
   delete          de-activate
   replace         replace
   destroy         destroy element and it's children

   navigating
   ----------
   getParent       (var par, el)     return true, par = parent false: 0
   getChild        (var child, el)   ..
   getNext         (var next, el)    ..
   getPrevious     (var prev, el)    ..
   hasParent       (el)              return true if parent
   hasChild        (el)                             child
   hasNext         (el)                             next
   hasPrevious     (el)                             previous

   nextChild       (var nextflag:boolean; var neCh:word) : boolean;

   scanElement     supply neighbour element in tree:
                   old link          search in priority order for
                   -------------     --------------------------------------
                   next,child        child,next,previous,parent
                   previous          previous, parent
                   parent            next,previous,parent

   --------------------------------------------------------------------------

   control
   ----
   inittree        clears links & undo counts
   undo            undo last operation
   startlinking    next action is linked
   stoplinking     next action is not linked

   global constants and functions in other unit:
   -----------------------------
   maxelement      total number of elements
   procedure treeMessage(const msg : TtreeNotification);
}

interface

uses sysutils;
                                                     
type TLink        = (ltNone,ltParent,ltChild,ltNext,ltPrevious);
     TUndomessage = (umDestroy,umUndoDelete,umUndoReplace);

//-- control

procedure initTree;         //initialize tree
procedure RemoveUndoCounts; //destroy deleted elements, remove undo counts
procedure startTreeLinking;
procedure stopTreeLinking;
procedure UndoLastStep;
function GetEdBase : byte;
function GetEdStep : byte;
procedure setEdBase(eb : byte);
procedure setEdstep(es : byte);
procedure addBackLinks;
function TestELdeleted(el : word) : boolean;

//--navigating

function hasParent(el: word) : boolean;
function hasChild(el:word) : boolean;
function hasNext(el: word) : boolean;
function hasPrevious(el: word) : boolean;
function getParent(var par : word; el : word) : boolean;
function getChild(var chld : word; el : word) : boolean;
function getNext(var nxt : word; el : word) : boolean;
function getPrevious(var prev : word; el : word) : boolean;
function ScanElement(var elmnt : word; lt : TLink) : TLink;
function fullScan(var elmnt : word; lt : Tlink) : Tlink;
function getChildCount(del : word) : word;
function getChildNr(chel : word) : word; //report rank 1.. of child 0:not found
function getLastChild(del : word) : word;
function nextChild(var nextflag : boolean; var neCh : word) : boolean;
function getChildELbyRank(pel : word; rank: byte; var ix : word) : word;
procedure GetChildN(var el : word; pel : word; n : word);

//--edit procedures

procedure PlaceA(del,sel : word);      //place element sel after del
procedure PlaceB(del,sel : word);      //place element sel before del
procedure PlaceChildA(del,sel : word); //place element sel as last child of del
procedure PlaceChildB(del,sel : word); //place element sel as 1st child of del
procedure deleteEL(del : word);        //delete element del
procedure replaceEL(del,sel : word);   //replace element del by sel
procedure destroyEl(del : word);       //destroy element del and children

implementation

uses unit1;

const UndoDepth = 15;

var
    startflag : boolean;    //set linkFlag on next step
    linkflag  : boolean;    //link edit actions (apply same stepNr)
    EdStep    : byte;       //edit action counter
    EdBase    : byte;       //1st action in circular counter

procedure destroyChildren(el : word);forward;

//--- hidden procedures ---

procedure clearLinks(el : word);
begin
 with element[el] do
  begin
   parent := 0;
   child := 0;
   next := 0;
   previous := 0;
   Cstep := 0;
   Dstep := 0;
  end;
end;

procedure copyLinks(var sel : word;del : word);
//copy links of element del to sel
begin
 with element[sel] do
  begin
   parent   := element[del].parent;
   child    := element[del].child;
   next     := element[del].next;
   previous := element[del].previous;
  end;
end;

procedure setParentLinks(par : word);
//set link of children to par
var chel :word;
begin
 chel := element[par].child;
 if chel > 0 then
  repeat
   element[chel].parent := par;
  until getNext(chel,chel) = false;
end;

procedure resetUndoBase;
//test element for Cstep=undoBase -> Cstep= 0
//destroy element if Dstep=undoBase
var el : word;
begin
 for el := 1 to maxElement do
  with element[el] do
   begin
    if (Cstep and $7f) = edBase then Cstep := 0;
    if (Dstep and $7f) = edBase then
     begin
      if (Dstep and $80) = 0 then destroyChildren(el);
      undomessage(umdestroy,0,el);
      destroyEL(el);
     end;
   end;
end;

procedure Extract(el : word);
//extract element el from tree, repair tree
//note: element[1] has no parent and cannot be extracted
//set child:=0 if not active link to children or replacement
var par,chd,nxt,prv : word;
begin
 with element[el] do
  begin
   par := parent;
   nxt := next;
   prv := previous;
   chd := child;
   if chd > 0 then
    begin
     if element[chd].parent <> el then child := 0; //brooken link
     if prv > 0 then
      if element[prv].child = chd then child := 0;//replaced link
    end;
  end;
 if nxt > 0 then element[nxt].previous := prv;
 if prv > 0 then element[prv].next := nxt
  else element[par].child := nxt;
end;

procedure LinkIn(el : word);
//element[el] has correct links
//parent,child,next,previous elements must link to el
var par,nxt,prv : word;
begin
 with element[el] do
  begin
   par := parent;
   nxt := next;
   prv := previous;
  end;
 if nxt > 0 then element[nxt].previous := el;
 if prv > 0 then element[prv].next := el
  else if par > 0 then element[par].child := el;
 setParentLinks(el);
end;

procedure NextStep;
//increment the undoBase/undoStep counter
begin
 if EdStep = 0 then
  begin
   EdStep := 1;
   EdBase := 1;
   linkFlag := startFlag;
   exit;
  end;
 if linkflag = false then begin
                           inc(EdStep);
                           linkFlag := startFlag;
                          end;
 if EdStep > UndoDepth then EdStep := 1;
 if EdStep = EdBase then begin
                          resetUndoBase;
                          inc(EdBase);
                          if EdBase > UndoDepth then EdBase := 1;
                         end;
end;

procedure previousStep;
//decrement undoBase/undoStep counter
begin
 if EdStep = 0 then exit;
//
if EdStep = EdBase then begin
                              EdStep := 0;
                              EdBase := 0;
                             end
  else if EdStep = 1 then EdStep := UndoDepth
        else dec(EdStep);
end;

//--- control ---

function TestELdeleted(el : word) : boolean;
//return true if element el or parent is deleted
begin
 repeat
  result := element[el].Dstep > 0;
 until result or (getParent(el,el)=false);
end;

function GetEdBase : byte;
begin
 result := EdBase;
end;

function GetEdStep : byte;
begin
 result := EdStep;
end;

procedure setEdBase(eb : byte);
begin
 edBase := eb;
end;

procedure setEdstep(es : byte);
begin
 edStep := es;
end;

procedure initTree;
//do not affect status and properties
var n : word;
begin
 for n:= 1 to maxElement do clearLinks(n);
 startFlag := false;
 linkflag := false;
 EdStep := 0;
 EdBase := 0;
end;

procedure startTreeLinking;
//start new edit commands
//clear linkflag, set startflag
begin
 if startFlag = false then
  begin
   linkflag := false;
   startFlag := true;
  end;
end;

procedure stopTreeLinking;
begin
 startflag := false;
 linkflag := false;
end;

//-- navigating --

function hasParent(el : word) : boolean;
begin
 result := element[el].parent > 0;  
end;

function hasChild(el:word) : boolean;
begin
 el := element[el].child;
 while (el > 0) and (element[el].Dstep > 0) do el := element[el].next;
 result := el > 0;
end;

function hasNext(el: word) : boolean;
begin
 repeat
  el := element[el].next;
  result := el > 0;
 until (result=false) or (element[el].Dstep= 0);
end;

function hasPrevious(el:word) : boolean;
begin
 repeat
  el := element[el].previous;
  result := el > 0;
 until (result=false) or (element[el].Dstep= 0);
end;

function getParent(var par:word; el:word) : boolean;
begin
 with element[el] do
  if parent > 0 then
   begin
    par := parent;
    result := true;
   end
   else result := false;
end;

function getChild(var chld : word; el: word) : boolean;
begin
 el := element[el].child;
 while (el > 0) and (element[el].Dstep > 0) do el := element[el].next;
 result := el > 0;
 if result then chld := el;
end;

function getNext(var nxt : word; el: word) : boolean;
begin
 repeat
  el := element[el].next;
 until (el=0) or (element[el].Dstep=0);
 result := el > 0;
 if result then nxt := el;
end;

function getPrevious(var prev:word; el:word) : boolean;
begin
 repeat
  el := element[el].previous;
 until (el=0) or (element[el].Dstep=0);
 result := el > 0;
 if result then prev := el;
end;

function getLastChild(del : word) : word;
//return 0 if no child
begin
 if getchild(del,del) then
  begin
   repeat until getnext(del,del) = false;
   result := del;
  end
  else result := 0;
end;

function getChildNr(chel : word) : word;
//return rank of child chel; 0: not found
var el,pel : word;
begin
 result := 0;
 if getparent(pel,chel) then
  begin
   getChild(el,pel);
   repeat
    inc(result);
   until (chel=el) or (getNext(el,el) = false);
  end;
end;

function getChildCount(del : word) : word;
//return number of children
begin
 result := 0;
 if getchild(del,del) then
  begin
   inc(result);
   while getnext(del,del) do inc(result);
  end;
end;

function getChildELbyRank(pel : word; rank: byte; var ix : word) : word;
//return child element nr:rank of el 0: no child
//if rank > element count then return highest element. ix is real rank
var el : word;
    GO : boolean;
begin
 ix := 0; result := 0;
 if getChild(el,pel) = false then exit;
//
 inc(ix);
 GO := true;
 while GO do
  begin
   result := el;
   if rank = ix then GO := false
    else GO := getNext(el,el);
   if GO then inc(ix);
  end;
end;

procedure GetChildN(var el : word; pel : word; n : word);
//entry: pel is macro
//exit:  el is child n
var i : word;
begin
 i := 1;
 getChild(el,pel);
 while i < n do begin
                 getNext(el,el);
                 inc(i);
                end;
end;

function nextChild(var nextflag : boolean; var neCh : word) : boolean;
//nextflag=false : return child
//        = true : return next
//return false if no child or next
//if false, neCh is unchanged
begin
 if nextflag then result := getnext(neCh,neCh)
  else begin
        result := getChild(neCh,neCh);
        nextflag := true;
       end;
end;

function ScanElement(var elmnt : word; lt : TLink) : TLink;
//walk thru tree, children first then next (depth 1st)
//replace elmnt by child,next,previous or parent element
begin
 result := ltNone;
 if (lt = ltChild) or (lt = ltNext) then
  begin
   if getchild(elmnt,elmnt) then result := ltChild
    else if getnext(elmnt,elmnt) then result := ltNext
          else if getprevious(elmnt,elmnt) then result := ltPrevious
           else if getparent(elmnt,elmnt) then result := ltParent;
  end;
 if result <> ltNone then exit;

 if lt = ltPrevious then
  begin
   if getprevious(elmnt,elmnt) then result := ltPrevious
    else if getparent(elmnt,elmnt) then result := ltParent;
  end;
 if result <> ltNone then exit;

 if lt = ltParent then
  begin
   if getnext(elmnt,elmnt) then result := ltNext
    else if getprevious(elmnt,elmnt) then result := ltPrevious
     else if getparent(elmnt,elmnt) then result:= ltParent;
  end;
end;

function fullScan(var elmnt : word; lt : Tlink) : Tlink;
//as scanelement but: deleted elements are not skipped
var chd,nxt,prv,par : word;
label childexit,nextexit,previousexit,parentexit,noneexit;
begin
 with element[elmnt] do
  begin
   chd := child;
   prv := previous;
   nxt := next;
   par := parent;
   case lt of
    ltChild,
    ltNext : begin
              if chd > 0 then goto childexit;
              if nxt > 0 then goto nextexit;
              if prv > 0 then goto previousexit;
              if par > 0 then goto parentexit;
             end;
    ltPrevious : begin
                  if prv > 0 then goto previousexit;
                  if par > 0 then goto parentexit;
                 end;
    ltParent : begin
                if nxt > 0 then goto nextexit;
                if prv > 0 then goto previousexit;
                if par > 0 then goto parentexit;
               end;
   end;//case
 end;
 result := ltNone;
 exit;

childexit:
  elmnt := chd;
  result := ltChild;
  exit;

nextexit:
  elmnt := nxt;
  result := ltNext;
  exit;

previousexit:
  elmnt := prv;
  result := ltPrevious;
  exit;

parentexit:
  elmnt := par;
  result := ltParent;

end;

procedure RemoveUndoCounts;
//destroy deleted elements
//set Edstep = 0; Cstep,Dstep of elements = 0;
var n : word;
begin
 for n := 1 to maxElement do
  with element[n] do
   begin
    Cstep := 0;
    if Dstep > 0 then
     begin
      if (Dstep and $80) = 0 then destroyChildren(n); //not replacement
      undoMessage(umDestroy,0,n);
      destroyEL(n);
     end;
   end;
 EdStep:= 0;
 EdBase := 0;
end;

//--edit actions

procedure PlaceA(del,sel : word);
//place free element sel after del
begin
 nextstep;
 with element[sel] do
  begin
   Cstep := EdStep;
   previous := del;
   next := element[del].next;
   parent := element[del].parent;
  end;
 LinkIn(sel);
end;

procedure PlaceB(del,sel : word);
//place free element sel before del
begin
 nextstep;
 with element[sel] do
  begin
   Cstep := EdStep;
   previous := element[del].previous;
   next := del;
   parent := element[del].parent;
  end;
 LinkIn(sel);
end;

procedure PlaceChildA(del,sel : word);
//place free element sel as last child of del
var chd : word;
begin
 nextstep;
 with element[sel] do
  begin
   parent := del;
   next := 0;
   Cstep := EdStep;
  end;
 chd := element[del].child;
 if chd > 0 then
  begin
   repeat until getnext(chd,chd) = false; //position on last child
   element[sel].previous := chd;
   element[chd].next := sel;
  end
  else begin
        element[sel].previous := 0;
        element[del].child := sel;
       end;
end;

procedure PlaceChildB(del,sel : word); //position as first child
//place free element sel as first child of del
var chd : word;
begin
 nextstep;
 chd := element[del].child;
 with element[sel] do
  begin
   previous := 0;
   next := chd;
   parent := del;
   Cstep := EdStep;
  end;
 element[del].child := sel;
 if chd > 0 then element[chd].previous := sel;
end;

procedure replaceEL(del,sel : word);
//replace element del by sel
begin
 nextstep;
 copylinks(sel,del);
 element[sel].previous := del;
 linkIn(sel);
 element[sel].Cstep := EdStep or $80;//add replace bit
 element[del].Dstep := EdStep or $80;//..
 element[del].child := 0;
end;

procedure deleteEL(del : word);
//delete element
//children are not deleted
begin
 nextStep;
 element[del].Dstep := EdStep;
end;

procedure destroyEL(del : word);
//does not report destruction
begin
 Extract(del);
 clearlinks(del);
end;

procedure destroyChildren(el : word);
var link : Tlink;
    oldEL : word;
    rootEL : word;
begin
 rootEL := el;
 if getChild(el,el) then
  begin
   link := ltNext;
   repeat
    oldEL := el;
    link := scanelement(el,link);
    if (link = ltParent) or (link = ltPrevious) then
     begin
      undoMessage(umDestroy,0,oldEL);
      clearLinks(oldEL);
     end;
   until (el = rootEL);
  end;
end;

//--- undo ---

procedure undoReplace(el : word);
//previous element restored, el destroyed, children of el linked to previous
var prv : word;
begin
 prv := element[el].previous;
 element[prv].child := element[el].child;
 setParentLinks(prv);
 undoMessage(umUndoReplace,prv,el);
 destroyEL(el);
 element[prv].Dstep := 0;
end;

procedure UndoDelete(el : word);
begin
 undoMessage(umUndoDelete,el,0);
 element[el].Dstep := 0;
end;

procedure UndoLastStep;
var chel,el,oldEL : word;
    link : Tlink;
begin
 if EdStep = 0 then exit;
//
 el := 1;
 link := ltChild;
 repeat
  oldEL := el;
  link := fullscan(el,link);
  if (link=ltParent) or (link=ltPrevious) then
   with element[oldEL] do
    begin
     if Dstep = EdStep then
      begin
       undoDelete(oldEL);
      end;
     if (Cstep and $7f) = edStep then
      if (Cstep and $80) = 0 then
       begin
        undomessage(umDestroy,0,oldEL);
        if getChild(chel,oldEL) then destroychildren(oldEL);
        destroyEL(oldEL);
       end
       else undoReplace(oldEL);
    end;
 until (link=ltNone);
 previousstep;
end;

procedure AddBackLinks;
//tree has only forward links: child,next
//add backward links: parent,previous
var el,oldEL : word;
    link : Tlink;
begin
 link := ltNext;
 el := 1;
 repeat
  oldEL := el;
  link := fullscan(el,link);
  if link = ltChild then setParentLinks(oldEL)
   else if link = ltnext then element[el].previous := oldEL;
 until (el=1) and ((link=ltNone) or (link=ltPrevious));
end;

end.





